@@ -1,4 +1,7 @@
 from __future__ import absolute_import
+
+from . import activations
+from . import applications
 from . import backend
 from . import datasets
 from . import engine
@@ -8,11 +11,11 @@ from . import utils
 from . import wrappers
 from . import callbacks
 from . import constraints
-from . import initializations
+from . import initializers
 from . import metrics
 from . import models
-from . import objectives
+from . import losses
 from . import optimizers
 from . import regularizers
 
-__version__ = '1.0.7'
+__version__ = '2.0.2'
