@@ -4,7 +4,5 @@ from .topology import InputSpec
 from .topology import Input
 from .topology import InputLayer
 from .topology import Layer
-from .topology import Merge
-from .topology import merge
 from .topology import get_source_inputs
 from .training import Model
